@@ -15,15 +15,35 @@ from cftool.misc import *
 from cfdata.tabular import *
 from tqdm import tqdm
 from functools import partial
-
+from trains import Task, Logger
 from abc import ABCMeta, abstractmethod
 
+try:
+    import torch.cuda.amp as amp
+except:
+    amp = None
+
+from .losses import *
 from .modules import *
 from .misc.toolkit import *
 
+trains_logger: Union[Logger, None] = None
 model_dict: Dict[str, Type["ModelBase"]] = {}
 
 
+class amp_autocast_context(context_error_handler):
+    def __init__(self, use_amp: bool):
+        self._autocast = None if not use_amp else amp.autocast()
+
+    def __enter__(self):
+        if self._autocast is not None:
+            self._autocast.__enter__()
+
+    def _normal_exit(self, exc_type, exc_val, exc_tb):
+        if self._autocast is not None:
+            self._autocast.__exit__(exc_type, exc_val, exc_tb)
+
+
 class SplitFeatures(NamedTuple):
     categorical: Union[torch.Tensor, tensor_dict_type, None]
     numerical: Union[torch.Tensor, None]
@@ -126,9 +146,11 @@ class ModelBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
     def _init_config(self,
                      tr_data: TabularData):
         self.tr_data = tr_data
+        self._loss_config = self.config.setdefault("loss_config", {})
         # TODO : optimize encodings by pre-calculate one-hot encodings in Wrapper
         self._encoding_methods = self.config.setdefault("encoding_methods", {})
         self._encoding_configs = self.config.setdefault("encoding_configs", {})
+        self._default_encoding_configs = self.config.setdefault("default_encoding_configs", {})
         self._default_encoding_method = self.config.setdefault("default_encoding_method", "embedding")
 
     def _init_loss(self,
@@ -136,11 +158,11 @@ class ModelBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
         if tr_data.is_reg:
             self.loss = nn.L1Loss()
         else:
-            self.loss = nn.CrossEntropyLoss()
+            self.loss = FocalLoss(self._loss_config)
 
     def _init_encoder(self, idx: int):
         methods = self._encoding_methods.setdefault(idx, self._default_encoding_method)
-        config = self._encoding_configs.setdefault(idx, {})
+        config = self._encoding_configs.setdefault(idx, self._default_encoding_configs)
         num_values = self.tr_data.recognizers[idx].num_unique_values
         if isinstance(methods, str):
             methods = [methods]
@@ -148,9 +170,14 @@ class ModelBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
         self.encoders[str(idx)] = EncoderStack(*encoders)
 
     def _optimizer_step(self,
-                        optimizers):
+                        optimizers,
+                        grad_scalar):
         for opt in optimizers.values():
-            opt.step()
+            if grad_scalar is None:
+                opt.step()
+            else:
+                grad_scalar.step(opt)
+                grad_scalar.update()
             opt.zero_grad()
 
     @staticmethod
@@ -213,14 +240,18 @@ class ModelBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
 class Pipeline(nn.Module, LoggingMixin):
     def __init__(self,
                  model: ModelBase,
+                 tracker: Tracker,
                  config: Dict[str, Any],
-                 verbose_level: int):
+                 verbose_level: int,
+                 is_loading: bool):
         super().__init__()
-        self._init_config(config)
+        self.tracker = tracker
+        self._init_config(config, is_loading)
         self.model = model
         self._verbose_level = verbose_level
+        self._no_grad_in_predict = True
 
-    def _init_config(self, config):
+    def _init_config(self, config, is_loading):
         self._wrapper_config = config
         self.config = config.setdefault("pipeline_config", {})
         self.batch_size = self.config.setdefault("batch_size", 128)
@@ -244,13 +275,16 @@ class Pipeline(nn.Module, LoggingMixin):
         else:
             self.ema_decay = EMA(ema_decay, self.model.named_parameters())
 
+        self._use_amp = config["use_amp"]
+        self.scaler = None if amp is None or not self._use_amp else amp.GradScaler()
+
         self._logging_path_ = config["_logging_path_"]
         self.logging_folder = config["logging_folder"]
         self.checkpoint_folder = self.config.setdefault(
             "checkpoint_folder",
             os.path.join(self.logging_folder, "checkpoints")
         )
-        if os.path.isdir(self.checkpoint_folder):
+        if not is_loading and os.path.isdir(self.checkpoint_folder):
             self.log_msg(
                 f"'{self.checkpoint_folder}' already exists, all of its contents will be removed",
                 self.warning_prefix, msg_level=logging.WARNING
@@ -425,7 +459,7 @@ class Pipeline(nn.Module, LoggingMixin):
             self.model.parameters(), self._clip_norm)
 
     def _optimizer_step(self):
-        self.model._optimizer_step(self.optimizers)
+        self.model._optimizer_step(self.optimizers, self.scaler)
 
     def _monitor_step(self):
         if self._step_count % self.num_step_per_snapshot == 0:
@@ -491,6 +525,16 @@ class Pipeline(nn.Module, LoggingMixin):
                 use_decayed = True
                 decayed_metrics[metric_type] = self.metrics_decay[metric_type].update("metric", sub_metric)
         metrics_for_scoring = decayed_metrics if use_decayed else metrics
+        if self.tracker is not None or trains_logger is not None:
+            for name, value in metrics_for_scoring.items():
+                if self.tracker is not None:
+                    self.tracker.track_scalar(name, value, iteration=self._step_count)
+                if trains_logger is not None:
+                    trains_logger.report_scalar(
+                        title="Evaluating",
+                        series=name, value=value,
+                        iteration=self._step_count
+                    )
         weighted_scores = [v * signs[k] * self.metrics_weights[k] for k, v in metrics_for_scoring.items()]
         score = sum(weighted_scores) / len(weighted_scores)
 
@@ -514,15 +558,27 @@ class Pipeline(nn.Module, LoggingMixin):
 
         return score, metrics
 
-    def _predict(self,
-                 loader: DataLoader,
-                 **kwargs) -> Dict[str, np.ndarray]:
-        with eval_context(self, no_grad=kwargs.get("no_grad", False)):
+    def _get_results(self,
+                     no_grad: bool,
+                     loader: DataLoader,
+                     **kwargs) -> Tuple[List[np.ndarray], List[tensor_dict_type]]:
+        with eval_context(self, no_grad=no_grad):
             results, labels = [], []
             for x_batch, y_batch in loader:
                 if y_batch is not None:
                     labels.append(y_batch)
                 results.append(self.model(self._collate_batch(x_batch, y_batch), **kwargs))
+        return labels, results
+
+    def _predict(self,
+                 loader: DataLoader,
+                 **kwargs) -> Dict[str, np.ndarray]:
+        no_grad = kwargs.pop("no_grad", self._no_grad_in_predict)
+        try:
+            labels, results = self._get_results(no_grad, loader, **kwargs)
+        except:
+            no_grad = self._no_grad_in_predict = False
+            labels, results = self._get_results(no_grad, loader, **kwargs)
         results = self.model._collate_tensor_dicts(results)
         results = {k: to_numpy(v) for k, v in results.items()}
         if labels:
@@ -565,12 +621,27 @@ class Pipeline(nn.Module, LoggingMixin):
                     self._step_count += 1
                     with timing_context(self, "collate batch"):
                         batch = self._collate_batch(x_batch, y_batch)
-                    with timing_context(self, "model.forward"):
-                        forward_results = self.model(batch)
-                    with timing_context(self, "loss.forward"):
-                        loss_dict = self.model.loss_function(batch, forward_results)
+                    with amp_autocast_context(self._use_amp):
+                        with timing_context(self, "model.forward"):
+                            forward_results = self.model(batch)
+                        with timing_context(self, "loss.forward"):
+                            loss_dict = self.model.loss_function(batch, forward_results)
+                    if self.tracker is not None or trains_logger is not None:
+                        for name, tensor in loss_dict.items():
+                            value = tensor.item()
+                            if self.tracker is not None:
+                                self.tracker.track_scalar(name, value, iteration=self._step_count)
+                            if trains_logger is not None:
+                                trains_logger.report_scalar(
+                                    "Training",
+                                    series=name, value=value,
+                                    iteration=self._step_count
+                                )
                     with timing_context(self, "loss.backward"):
-                        loss_dict["loss"].backward()
+                        loss = loss_dict["loss"]
+                        if self._use_amp:
+                            loss = self.scaler.scale(loss)
+                        loss.backward()
                     if self._clip_norm > 0.:
                         with timing_context(self, "clip_norm_step"):
                             self._clip_norm_step()
@@ -654,10 +725,14 @@ class Wrapper(LoggingMixin):
                  config: Union[str, Dict[str, Any]] = None,
                  *,
                  increment_config: Union[str, Dict[str, Any]] = None,
+                 tracker_config: Dict[str, Any] = None,
                  cuda: Union[str, int] = None,
                  verbose_level: int = 2):
+        self.tracker = None if tracker_config is None else Tracker(**tracker_config)
         self._verbose_level = int(verbose_level)
-        if cuda is not None:
+        if cuda == "cpu":
+            self.device = torch.device("cpu")
+        elif cuda is not None:
             self.device = torch.device(f"cuda:{cuda}")
         else:
             self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
@@ -665,6 +740,11 @@ class Wrapper(LoggingMixin):
         update_dict(increment_config, self.config)
         self._init_config()
 
+    def __str__(self):
+        return f"{type(self.model).__name__}()"
+
+    __repr__ = __str__
+
     @property
     def train_set(self) -> TabularDataset:
         raw = self.tr_data.raw
@@ -699,6 +779,7 @@ class Wrapper(LoggingMixin):
         self._binary_metric = self.config.setdefault("binary_metric", "acc")
         self._is_binary = self.config.get("is_binary")
         self._binary_threshold = self.config.get("binary_threshold")
+        self.config.setdefault("use_amp", False)
         logging_folder = self.config["logging_folder"] = self.config.setdefault(
             "logging_folder",
             os.path.join("_logging", model_dict[self._model].__identifier__)
@@ -711,38 +792,43 @@ class Wrapper(LoggingMixin):
         self.config["_logging_path_"] = logging_path
         self._init_logging(self._verbose_level, self.config.setdefault("trigger_logging", False))
 
-    def _prepare_modules(self):
+    def _prepare_modules(self,
+                         *,
+                         is_loading: bool = False):
         # model
         with timing_context(self, "init model"):
             self.model = model_dict[self._model](self.config, self.tr_data, self.device)
         # pipeline
         with timing_context(self, "init pipeline"):
-            self.pipeline = Pipeline(self.model, self.config, self._verbose_level)
+            self.pipeline = Pipeline(self.model, self.tracker, self.config, self._verbose_level, is_loading)
         # to device
         with timing_context(self, "init device"):
             self.pipeline.to(self.device)
 
-    # api
-
-    def fit(self,
-            x: data_type,
-            y: data_type = None,
-            x_cv: data_type = None,
-            y_cv: data_type = None) -> "Wrapper":
+    def _before_loop(self,
+                     x: data_type,
+                     y: data_type = None,
+                     x_cv: data_type = None,
+                     y_cv: data_type = None):
         # data
         y, y_cv = map(to_2d, [y, y_cv])
         args = (x, y) if y is not None else (x,)
         self._data_config["verbose_level"] = self._verbose_level
-        self.tr_data = TabularData(**self._data_config).read(*args, **self._read_config)
+        self._original_data = TabularData(**self._data_config).read(*args, **self._read_config)
+        self.tr_data = self._original_data
+        self._save_original_data = False
         if x_cv is not None:
             self.cv_data = self.tr_data.copy_to(x_cv, y_cv)
         else:
             if self._cv_split <= 0.:
                 self.cv_data = None
             else:
+                self._save_original_data = True
                 self.cv_data, self.tr_data = self.tr_data.split(self._cv_split)
         # modules
         self._prepare_modules()
+
+    def _loop(self):
         # training loop
         self.pipeline(self.tr_data, self.cv_data)
         # binary threshold
@@ -761,6 +847,58 @@ class Wrapper(LoggingMixin):
                     self._binary_threshold = None
         # logging
         self.log_timing()
+
+    # api
+
+    def fit(self,
+            x: data_type,
+            y: data_type = None,
+            x_cv: data_type = None,
+            y_cv: data_type = None) -> "Wrapper":
+        self._before_loop(x, y, x_cv, y_cv)
+        self._loop()
+        return self
+
+    def trains(self,
+               x: data_type,
+               y: data_type = None,
+               x_cv: data_type = None,
+               y_cv: data_type = None,
+               *,
+               trains_config: Dict[str, Any] = None,
+               keep_task_open: bool = False,
+               queue: str = None) -> "Wrapper":
+        if trains_config is None:
+            return self.fit(x, y, x_cv, y_cv)
+        # init trains
+        if trains_config is None:
+            trains_config = {}
+        project_name = trains_config.get("project_name")
+        task_name = trains_config.get("task_name")
+        if queue is None:
+            task = Task.init(**trains_config)
+            cloned_task = None
+        else:
+            task = Task.get_task(project_name=project_name, task_name=task_name)
+            cloned_task = Task.clone(source_task=task, parent=task.id)
+        # before loop
+        self._verbose_level = 6
+        self._data_config["verbose_level"] = 6
+        self._before_loop(x, y, x_cv, y_cv)
+        self.pipeline.use_tqdm = False
+        copied_config = shallow_copy_dict(self.config)
+        if queue is not None:
+            cloned_task.set_parameters(copied_config)
+            Task.enqueue(cloned_task.id, queue)
+            return self
+        # loop
+        task.connect(copied_config)
+        global trains_logger
+        trains_logger = task.get_logger()
+        self._loop()
+        if not keep_task_open:
+            task.close()
+            trains_logger = None
         return self
 
     def predict(self,
@@ -792,6 +930,15 @@ class Wrapper(LoggingMixin):
         raw = self.pipeline.predict(x, **kwargs)
         return self.model.to_prob(raw)
 
+    def to_pattern(self, **kwargs) -> ModelPattern:
+        def _predict(x):
+            return self.predict(x, **kwargs)
+
+        def _predict_prob(x):
+            return self.predict_prob(x, **kwargs)
+
+        return ModelPattern(predict_method=_predict, predict_prob_method=_predict_prob)
+
     def save(self,
              folder: str = None,
              *,
@@ -802,9 +949,13 @@ class Wrapper(LoggingMixin):
         base_folder = os.path.dirname(abs_folder)
         with lock_manager(base_folder, [folder]):
             Saving.prepare_folder(self, folder)
-            self.tr_data.save(os.path.join(folder, "__data__", "train"), compress=compress)
-            if self.cv_data is not None:
-                self.cv_data.save(os.path.join(folder, "__data__", "valid"), compress=compress)
+            train_data_folder = os.path.join(folder, "__data__", "train")
+            if self._save_original_data:
+                self._original_data.save(train_data_folder, compress=compress)
+            else:
+                self.tr_data.save(train_data_folder, compress=compress)
+                if self.cv_data is not None:
+                    self.cv_data.save(os.path.join(folder, "__data__", "valid"), compress=compress)
             self.pipeline.save_checkpoint(folder)
             self.config["is_binary"] = self._is_binary
             self.config["binary_threshold"] = self._binary_threshold
@@ -831,7 +982,7 @@ class Wrapper(LoggingMixin):
                 cv_data = None
                 if os.path.isdir(cv_data_folder) or os.path.isfile(f"{cv_data_folder}.zip"):
                     cv_data = wrapper.cv_data = TabularData.load(cv_data_folder, compress=compress)
-                wrapper._prepare_modules()
+                wrapper._prepare_modules(is_loading=True)
                 pipeline = wrapper.pipeline
                 pipeline.restore_checkpoint(folder)
                 pipeline._init_data(tr_data, cv_data)
